// This is a personal academic project. Dear PVS-Studio, please check it.
// PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <signal.h>
#include <stdbool.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <fcntl.h>
#include <getopt.h>

#define MAX_SIZE 524
#define MAX_RECV_SIZE 1024
#define N(x) (sizeof(x)/sizeof((x)[0]))

static void version(void) {
    fprintf(stderr, "BProxy v1.0\nITBA Protocolos de Comunicación 2021/1 -- Grupo 7\nVer LICENSE.md\n");
}

static void usage(const char *progname) {
    fprintf(stderr,
        "Usage: %s [token] [OPTION]\n\n"
        "   -h                     Imprime la ayuda y termina.\n"
        "   -m                     Obtener métricas del servidor.\n"
        "   -b                     Obtener el tamaño del buffer.\n"
        "   -B <value>             Modificar el tamaño del buffer.\n"
        "   -t                     Obtener el valor del timeout.\n"
        "   -T <value>             Modificar el valor del timeout.\n"
        "   -f                     Obtener cantidad de páginas de nombres de usuario.\n"
        "   -u <page>              Obtener nombres de usuario de la página especificada.\n"
        "   -l <name>              Obtener la última conexión del usuario específicado.\n"
        "   -c <name>:<new_name>   Nombre de usuario y nuevo nombre de usuario\n"
        "   -p <name>:<new_pass>   Nombre de usuario y su nueva contraseña\n"
        "   -n <name>:<pass>       Nombre de usuario y contraseña del usuario que desea ser agregado\n"
        "   -d <name>              Nombre del usuario que desea ser borrado\n"
        "   -s                     Obtener el estado del disector de contraseñas POP3.\n"
        "   -S <estado>            Modificar el estado del disector de contraseñas POP3.\n"
        "   -a                     Obtener el estado de la autorización del proxy.\n"
        "   -A <estado>            Modificar el estado de la autorización del proxy.\n"
        "   -z                     Obtener el estado del servidor proxy.\n"
        "   -Z <estado>            Modificar el estado del servidor proxy.\n"
        "   -r                     Reinicia el servidor\n\n",
        progname);
    exit(1);
}

void parse_args(int argc, char ** argv, uint8_t * buffer) {
    int c;

    if (argc <= 1 || argv[1][0] == '-') {
        fprintf(stderr, "You must enter a valid token.\n");
        exit(EXIT_FAILURE);
    }
    
    uint64_t token = atol(argv[1]);
    
    buffer[0] = 0x01;

    buffer[1] = token>>56;
    buffer[2] = (uint8_t) (token>>48);
    buffer[3] = (uint8_t) (token>>40);
    buffer[4] = (uint8_t) (token>>32);
    buffer[5] = (uint8_t) (token>>24);
    buffer[6] = (uint8_t) (token>>16);
    buffer[7] = (uint8_t) (token>>8);
    buffer[8] = (uint8_t) token;

    for(int i=1; i<argc; ++i)
        argv[i] = argv[i+1];
    argc--;

    c = getopt(argc, argv, "hmbtfsazrB:T:l:u:c:p:n:d:S:A:Z:");
    // retorna : o ? si le falta el argumento
    if (c == -1)
        return;

    if (optind > 3 || argc == 2) {
        fprintf(stderr, "Only one option allowed\n");
        exit(EXIT_FAILURE);
    }
    
    uint16_t value;
    uint8_t ulen, plen, nulen;
    char * p;
    switch (c) {
        case 'h':
            usage(argv[0]);
            break;
        case 'm':
            buffer[9] = 0x00;
            break;
        case 'b':
            buffer[9] = 0x01;
            break;
        case 'B':
            buffer[9] = 0x02;
            buffer[9] = (uint16_t) atoi(optarg);
            break;
        case 't':
            buffer[9] = 0x03;
            break;
        case 'T':
            buffer[9] = 0x04;
            value = atoi(optarg);
            buffer[10] = value >> 8;
            buffer[11] = (uint8_t) value;
            break;
        case 'f':
            buffer[9] = 0x05;
            break;
        case 'u':
            buffer[9] = 0x06;
            value = atoi(optarg);
            buffer[10] = value >> 8;
            buffer[11] = (uint8_t) value;
            break;
        case 'l':
            buffer[9] = 0x07;
            ulen = strlen(optarg);
            buffer[10] = ulen;
            strcat((char *) buffer, optarg);
            break;
        case 'c':
            buffer[9] = 0x08;
            char *n = strchr(optarg, ':');
            *n = 0;
            n++;
            ulen = strlen(optarg);
            nulen = strlen(n);
            buffer[10] = ulen;
            strcat((char *) buffer, optarg);
            buffer[11+ulen] = nulen;
            strcat((char *) buffer, n);
            break;
        case 'p':
            buffer[9] = 0x09;
            p = strchr(optarg, ':');
            *p = 0;
            p++;
            ulen = strlen(optarg);
            plen = strlen(p);
            buffer[10] = ulen;
            strcat((char *) buffer, optarg);
            buffer[11+ulen] = plen;
            strcat((char *) buffer, p);
            break;
        case 'n':
            buffer[9] = 0x0A;
            p = strchr(optarg, ':');
            *p = 0;
            p++;
            ulen = strlen(optarg);
            plen = strlen(p);
            buffer[10] = ulen;
            strcat((char *) buffer, optarg);
            buffer[11+ulen] = plen;
            strcat((char *) buffer, p);
            break;
        case 'd':
            buffer[9] = 0x0B;
            ulen = strlen(optarg);
            buffer[10] = ulen;
            strcat((char *) buffer, optarg);
            break;
        case 's':
            buffer[9] = 0x0C;
            break;
        case 'S':
            buffer[9] = 0x0D;
            buffer[10] = (uint8_t) atoi(optarg);
            break;
        case 'a':
            buffer[9] = 0x0E;
            break;
        case 'A':
            buffer[9] = 0x0F;
            buffer[10] = (uint8_t) atoi(optarg);
            break;
        case 'z':
            buffer[9] = 0x10;
            break;
        case 'Z':
            buffer[9] = 0x11;
            buffer[10] = (uint8_t) atoi(optarg);
            break;
        case 'r':
            buffer[9] = 0x12;
            break;
        case 'v':
            version();
            exit(EXIT_SUCCESS);
        default:
            fprintf(stderr, "Unknown argument %d.\n", c);
            exit(EXIT_FAILURE);
    }

    // if (optind < argc + 1) {
    //     fprintf(stderr, "Argument not accepted: ");
    //     while (optind < argc) {
    //         fprintf(stderr, "%s ", argv[optind++]);
    //     }
    //     fprintf(stderr, "\n");
    //     exit(1);
    // }
}


int main(int argc, char **argv) {
    const char *err_msg = NULL;

    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    addr.sin_port = htons(8080);

    const int client = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (client < 0) {
        err_msg = "Unable to create UDP socket";
        goto finally;
    }

    if (connect(client, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
        err_msg = "Unable to connect to UDP socket";
        goto finally;
    }

    uint8_t buffer[MAX_SIZE] = {0};
    parse_args(argc, argv, buffer);
    for (int i = 0; i < 100; i++) {
        fprintf(stderr, "%x ", buffer[i]);
    }
    fprintf(stderr, "\n");

    
    ssize_t sent = sendto(client, buffer, N(buffer), 0, NULL, sizeof(addr));
    if (sent < 0) {
        err_msg = "Unable to send cmd";
        goto finally;
    }
    char * recv_buffer = malloc(MAX_RECV_SIZE);
    ssize_t received = recvfrom(client, recv_buffer, sizeof(recv_buffer), 0, NULL, NULL);
    if (received < 0) {
        err_msg = "Unable to receive cmd";
        goto finally;
    }

    printf("%s\n", recv_buffer);

    err_msg = NULL;

    int ret = EXIT_SUCCESS;
finally:
    free(recv_buffer);
    if (err_msg) {
        perror(err_msg);
        ret = EXIT_FAILURE;
    }

    if (client >= 0) {
        close(client);
    }
    return ret;
}
