// This is a personal academic project. Dear PVS-Studio, please check it.
// PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
#include <stdio.h>
#include <string.h>
#include <arpa/inet.h>

#include "buffer.h"
#include "selector.h"
#include "confignio.h"
#include "cmd.h"

#define N(x) (sizeof(x)/sizeof((x)[0]))

#define TOKEN 0x0FF1CEDEADB00B1E

//enum config_state {
//    READ,
//    WRITE,
//    DONE,
//    ERROR,
//};

static void on_token(struct cmd_parser * p, const uint64_t token) {
    *p->verified = (token - (uint64_t) TOKEN == 0);
    p->state = cmd_cmd;
    if (!*p->verified)
        p->state = cmd_error_invalid_token;
}

static void on_cmd(struct cmd_parser * p, const uint8_t cmd) {
    uint8_t * selected = p->cmd;
    switch (cmd) {
        case CMD_GET_METRICS:
            * selected = CMD_GET_METRICS;
            break;
        case CMD_GET_BUFFER_SIZE:
            * selected = CMD_GET_BUFFER_SIZE;
            break;
        case CMD_SET_BUFFER_SIZE:
            * selected = CMD_SET_BUFFER_SIZE;
            p->state = cmd_parameters;
            p->remaining = 2;
            p->param_state = param_value;
            break;
        case CMD_GET_TIMEOUT:
            * selected = CMD_GET_TIMEOUT;
            break;
        case CMD_SET_TIMEOUT:
            * selected = CMD_SET_TIMEOUT;
            p->state = cmd_parameters;
            p->remaining = 2;
            p->param_state = param_value;
            break;
        case CMD_GET_USER_PAGES:
            * selected = CMD_GET_USER_PAGES;
            break;
        case CMD_LIST_USERS:
            * selected = CMD_LIST_USERS;
            p->state = cmd_parameters;
            p->param_state = param_byte;
            p->remaining = 1;
            break;
        case CMD_GET_USER_LAST_CONNECTION:
            * selected = CMD_GET_USER_LAST_CONNECTION;
            p->state = cmd_parameters;
            p->param_state = param_ulen;
            p->remaining = 1;
            break;
        case CMD_MODIFY_USERNAME:
            * selected = CMD_MODIFY_USERNAME;
            p->state = cmd_parameters;
            p->param_state = param_ulen;
            p->remaining = 1;
            break;
        case CMD_MODIFY_PASSWORD:
            * selected = CMD_MODIFY_PASSWORD;
            p->state = cmd_parameters;
            p->param_state = param_ulen;
            p->remaining = 1;
            break;
        case CMD_ADD_USER:
            * selected = CMD_ADD_USER;
            p->state = cmd_parameters;
            p->param_state = param_ulen;
            p->remaining = 1;
            break;
        case CMD_DELETE_USER:
            * selected = CMD_DELETE_USER ;
            p->state = cmd_parameters;
            p->param_state = param_ulen;
            p->remaining = 1;
            break;
        case CMD_GET_PASSWORD_DISSECTOR_STATUS:
            * selected = CMD_GET_PASSWORD_DISSECTOR_STATUS;
            break;
        case CMD_MODIFY_PASSWORD_DISSECTOR_STATUS:
            * selected = CMD_MODIFY_PASSWORD_DISSECTOR_STATUS;
            * selected = CMD_LIST_USERS;
            p->state = cmd_parameters;
            p->param_state = param_byte;
            p->remaining = 1;
            break;
        case CMD_GET_AUTHENTICATION_STATUS:
            * selected = CMD_GET_AUTHENTICATION_STATUS;
            break;
        case CMD_MODIFY_AUTHENTICATION_STATUS:
            * selected = CMD_MODIFY_AUTHENTICATION_STATUS;
            * selected = CMD_LIST_USERS;
            p->state = cmd_parameters;
            p->param_state = param_byte;
            p->remaining = 1;
            break;
        case CMD_GET_PROXY_SERVER_STATUS:
            * selected = CMD_GET_PROXY_SERVER_STATUS;
            break;
        case CMD_CHANGE_PROXY_SERVER_STATUS:
            * selected = CMD_CHANGE_PROXY_SERVER_STATUS;
            * selected = CMD_LIST_USERS;
            p->state = cmd_parameters;
            p->param_state = param_byte;
            p->remaining = 1;
            break;
        case CMD_RESTART_PROXY_SERVER:
            * selected = CMD_RESTART_PROXY_SERVER;
            break;
        default:
            * selected = CMD_NOT_SUPPORTED_CMD;
    }
}

void cmd_process(struct cmd_st * d) {
//    unsigned ret = WRITE;

    uint8_t m = d->status;
    const uint8_t r = m ? 0x00 : 0xFF;
    if (cmd_marshall(d->wb, (enum cmd_response_status) r) == -1) {
//        ret  = ERROR;
    }

//    return ret;
}

#define ATTACHMENT(key) ( (struct config *)(key)->data)

static void config_read_init(struct selector_key * key) {
    struct cmd_st *d = &ATTACHMENT(key)->client.cmd;

    d->rb = &(ATTACHMENT(key)->read_buffer);
    d->wb = &(ATTACHMENT(key)->write_buffer);
    d->parser.cmd = &d->cmd;
    d->parser.verified = &d->verified;
    d->parser.on_cmd = on_cmd;
    d->parser.on_token = on_token;
    d->parser.idx = 0;
    cmd_parser_init(&d->parser);
}

void config_read(struct selector_key *key) {
    struct cmd_st *d = &ATTACHMENT(key)->client.cmd;
    bool error = false;
    uint8_t *ptr;
    size_t  count;
    ssize_t  n;
    struct sockaddr_storage client_addr;
    socklen_t client_addr_len = sizeof(client_addr);

    config_read_init(key);
    ptr = buffer_write_ptr(d->rb, &count);
    n = recvfrom(key->fd, ptr, BUFF_SIZE, 0, (struct sockaddr *) &client_addr, &client_addr_len);
    if (n > 0) {
        buffer_write_adv(d->rb, n);
        const enum cmd_state st = cmd_consume(d->rb, &d->parser, &error);

        if (cmd_is_done(st, 0)) {
            if (SELECTOR_SUCCESS == selector_set_interest_key(key, OP_WRITE)) {
                cmd_process(d);
            } else {
                return;
            }
        }
    } else {
        return;
    }
}

void config_write(struct selector_key *key) {
    struct cmd_st *d = &ATTACHMENT(key)->client.cmd;
    uint8_t *ptr;
    size_t  count;
    ssize_t  n;
    struct sockaddr_storage client_addr;
    socklen_t client_addr_len = sizeof(client_addr);

    ptr = buffer_read_ptr(d->wb, &count);
    n = sendto(key->fd, ptr, count, 0, (struct sockaddr *) &client_addr, sizeof(client_addr_len));
    if (n == -1) {
        return;
    } else {
        buffer_read_adv(d->wb, n);
        if (!buffer_can_read(d->wb)) {
            if (SELECTOR_SUCCESS != selector_set_interest_key(key, OP_READ)) {
                return;
            }
        }
    }
}
